#import "AudioFileInputPatch.h"
#import "QCStructure.h"
#import "QCImage.h"
#import "QCImagePixelBuffer.h"

@implementation AudioFileInputPatch : QCPatch

+ (int)executionModeWithIdentifier:(id)fp8
{
	return 2;
}

+ (BOOL)allowsSubpatchesWithIdentifier:(id)fp8
{
	return NO;
}

+ (int)timeModeWithIdentifier:(id)fp8
{
	return 1;
}

- (id)initWithIdentifier:(id)fp8
{
	if(self=[super initWithIdentifier:fp8])
	{
	}
	
	return self;
}

- (id)attributes
{
	NSMutableDictionary *at=[[super attributes] mutableCopy];
	[at setObject:@"Kineme Audio File Input" forKey:@"name"];
	return at;
}

- (BOOL)setup:(QCOpenGLContext *)context
{
	// See http://lists.apple.com/archives/Coreaudio-api/2006/Nov/msg00116.html for 64-bit insanity on this
	mBufferList = (AudioBufferList*)calloc(1,sizeof(AudioBufferList));

	return YES;
}
- (void)cleanup:(QCOpenGLContext *)context
{
	if(audioFile)
		ExtAudioFileDispose(audioFile);
	if(mBufferList->mBuffers[0].mData)
		free(mBufferList->mBuffers[0].mData);
	free(mBufferList);
}

- (void)enable:(QCOpenGLContext *)context
{
	cs = CGColorSpaceCreateWithName(kCGColorSpaceGenericRGB); 
}
- (void)disable:(QCOpenGLContext *)context
{
	CGColorSpaceRelease(cs);
}

- (BOOL)execute:(QCOpenGLContext *)context time:(double)time arguments:(NSDictionary *)arguments
{
	if([inputPath wasUpdated])
	{
		NSString *path = [inputPath stringValue];
		if([path length] > 0)
		{
			if([path characterAtIndex:0] != '/')
			{
				if([path characterAtIndex:0] != '~')
				{
					// not absolute and not ~-ified, must be composition relative
					NSDictionary *options = [self _getUserInfo:@".options"];
					NSURL *url = [options objectForKey:@"compositionURL"];
					path = [[[url path] stringByDeletingLastPathComponent] stringByAppendingPathComponent:path];
				}
				else    // it is a ~ directory, expand it
					path = [path stringByExpandingTildeInPath];
			}
		}
		NSURL *url = [NSURL fileURLWithPath: path];
		if(mBufferList->mBuffers[0].mData)
		{
			free(mBufferList->mBuffers[0].mData);
			mBufferList->mBuffers[0].mData = NULL;
		}
		if(audioFile)
			ExtAudioFileDispose(audioFile);
		if(ExtAudioFileOpenURL((CFURLRef)url, &audioFile))
		{
			[self logMessage: [NSString stringWithFormat:@"Error opening '%@'", [inputPath stringValue]]];
			audioFile = NULL;
			return YES;
		}
		AudioStreamBasicDescription fileAsbd, clientAsbd;
		UInt32 size = sizeof(fileAsbd);
		memset(&fileAsbd, 0, size);
		memset(&clientAsbd, 0, size);
		if(ExtAudioFileGetProperty(audioFile, kExtAudioFileProperty_FileDataFormat, &size, &fileAsbd))
		{
			[self logMessage: @"Error reading audio file properties!"];
			ExtAudioFileDispose(audioFile);
			audioFile = NULL;
			return YES;
		}
		
		/*NSLog(@"asbd infos: (size = %i vs. %i)", size, sizeof(fileAsbd));
		
		NSLog(@"   * Sample Rate:   %8f", fileAsbd.mSampleRate);
		NSLog(@"   * FormatID:      %i", fileAsbd.mFormatID);
		NSLog(@"   * FormatFlags:   %x", fileAsbd.mFormatFlags);
		NSLog(@"   * b per packet:  %i", fileAsbd.mBytesPerPacket);
		NSLog(@"   * f per packet:  %i", fileAsbd.mFramesPerPacket);
		NSLog(@"   * b per frame:   %i", fileAsbd.mBytesPerFrame);
		NSLog(@"   * channels:      %i", fileAsbd.mChannelsPerFrame);
		NSLog(@"   * bit depth:     %i", fileAsbd.mBitsPerChannel);*/
		mBufferList->mBuffers[0].mNumberChannels = 2;
		mBufferList->mBuffers[0].mDataByteSize = 512*4*fileAsbd.mChannelsPerFrame;
		mBufferList->mBuffers[0].mData = malloc(512*4*fileAsbd.mChannelsPerFrame);
		mBufferList->mBuffers[0].mNumberChannels = fileAsbd.mChannelsPerFrame;
		
		sampleRate = fileAsbd.mSampleRate;
		clientAsbd.mSampleRate = fileAsbd.mSampleRate;
		clientAsbd.mFormatFlags = kAudioFormatFlagIsFloat | kAudioFormatFlagIsPacked;
		clientAsbd.mFormatID = kAudioFormatLinearPCM;
		clientAsbd.mChannelsPerFrame = fileAsbd.mChannelsPerFrame;
		clientAsbd.mFramesPerPacket = 1;
		clientAsbd.mBytesPerFrame = sizeof(float) * fileAsbd.mChannelsPerFrame;
		clientAsbd.mBytesPerPacket = clientAsbd.mBytesPerFrame;
		clientAsbd.mBitsPerChannel = sizeof(float) * 8;
		if(fileAsbd.mFormatID == kAudioFormatMPEGLayer3)
			isMP3 = YES;
		else
			isMP3 = NO;
		
		if(ExtAudioFileSetProperty(audioFile, kExtAudioFileProperty_ClientDataFormat, sizeof(clientAsbd), &clientAsbd))
		{
			[self logMessage: @"Error setting client properties!"];
			ExtAudioFileDispose(audioFile);
			audioFile = NULL;
			return YES;
		}
		AudioConverterRef acRef;
		UInt32 acrsize=sizeof(AudioConverterRef);
		if(ExtAudioFileGetProperty(audioFile, kExtAudioFileProperty_AudioConverter, &acrsize, &acRef))
		{
			[self logMessage: @"Error getting audio converter!"];
			ExtAudioFileDispose(audioFile);
			audioFile = NULL;
			return YES;
		}
		
		AudioConverterPrimeInfo primeInfo;
		UInt32 piSize=sizeof(AudioConverterPrimeInfo);
		if(AudioConverterGetProperty(acRef, kAudioConverterPrimeInfo, &piSize, &primeInfo) != kAudioConverterErr_PropertyNotSupported)
			headerFrames = primeInfo.leadingFrames;
		else
			headerFrames = 0;
	}

	if(audioFile)
	{
		UInt32 numFrames = 512;
		mBufferList->mNumberBuffers = 1;
		mBufferList->mBuffers[0].mDataByteSize = 512*4 * mBufferList->mBuffers[0].mNumberChannels;
		// clear so if we go over we get proper zeros at the end if we have a partial packet
		memset(mBufferList->mBuffers[0].mData, 0, mBufferList->mBuffers[0].mNumberChannels * 512 * sizeof(float));

		
		if(isMP3)
		{
			// MP3 is dumb, and produces silence for the first few hundred samples after a seek.
			// to work around that, we seek to 1024 samples before our desired point, and decode twice to catch up
			// with a populated decoder stream.
			if(ExtAudioFileSeek(audioFile, MAX(time * sampleRate + headerFrames - 1024, headerFrames)) )
			{
				[self logMessage: @"Error seeking to current frame...!"];
				return YES;
			}
			ExtAudioFileRead(audioFile, &numFrames, mBufferList);
			numFrames = 512;
			ExtAudioFileRead(audioFile, &numFrames, mBufferList);
			numFrames = 512;
		}
		else
		{
			if(ExtAudioFileSeek(audioFile, time * sampleRate + headerFrames))
			{
				[self logMessage: @"Error seeking to current frame...!"];
				return YES;
			}
		}
		
		if(ExtAudioFileRead(audioFile, &numFrames, mBufferList) || numFrames != 512)
		{
			// if we can't read all 512 samples (EOF or error), manually set our buffer size stuff
			mBufferList->mBuffers[0].mDataByteSize = mBufferList->mBuffers[0].mNumberChannels * 512 * sizeof(float);
		}
		QCStructure *d = [[QCStructure alloc] init], *chan;
		QCStructure *p = [[QCStructure alloc] init];
		unsigned int i, j;
		NSUInteger currentChannel=0;
		//for(currentBuffer=0;currentBuffer<audioBufferList->mNumberBuffers;++currentBuffer)
		{
			AudioBuffer audioBuffer = mBufferList->mBuffers[0];
			
			float *sampleData = (float*)audioBuffer.mData;
			float max;
			unsigned int dataSize = audioBuffer.mDataByteSize/(audioBuffer.mNumberChannels * sizeof(float));

			for(j = 0; j < audioBuffer.mNumberChannels; ++j)
			{
				NSMutableArray *channel = [(NSMutableArray*)[NSMutableArray alloc] initWithCapacity: dataSize];
				unsigned int offset = j;
				max  = fabs(sampleData[offset]);
				for(i=0; i< dataSize; ++i)
				{
					[channel addObject: [NSNumber numberWithFloat: sampleData[offset]]];
					if( fabs(sampleData[offset]) > max)
						max = fabs(sampleData[offset]);
					offset += audioBuffer.mNumberChannels;
				}
				chan = [[QCStructure alloc] initWithArray: channel];
				[channel release];
				NSString *channelNumber = [NSString stringWithFormat:@"channel%02i",currentChannel];
				[d addMember: chan forKey: channelNumber];
				[p addMember: [NSNumber numberWithFloat: max] forKey: channelNumber];
				++currentChannel;
				[chan release];
			}
		}
		[outputWaveform setStructureValue: d];
		[outputPeaks setStructureValue: p];
		[d release];
		[p release];

		QCImagePixelBuffer *pb = [[context imageManager] createPixelBufferWithFormat: [QCPixelFormat pixelFormatARGB8]
																		  pixelsWide: 512
																		  pixelsHigh: mBufferList->mBuffers[0].mNumberChannels
																			 options: nil];
		
		unsigned int *audioImageData = (unsigned int*)[pb baseAddress];
		unsigned int rowBytes = [pb bytesPerRow]/sizeof(unsigned int);

		[pb beginUpdatePixels: FALSE colorSpace: cs];
		for(j=0;j<mBufferList->mBuffers[0].mNumberChannels;++j)
		{
			for(i=0;i<512;++i)
			{
				unsigned int value = (127.f * (1.0f + ((float*)mBufferList->mBuffers[0].mData)[i*mBufferList->mBuffers[0].mNumberChannels + j]));
				value *= 0x01010101;
				audioImageData[i] = value;
			}
			audioImageData += rowBytes;
		}
		[pb endUpdatePixels];
		QCImage *audioImage = [[QCImage alloc] initWithQCImageBuffer: pb options: nil];
		[pb release];
		[outputWaveformImage setImageValue: audioImage];
		[audioImage release];
	}
	
	return YES;
}

@end
