#import "AudioInputPatch.h"
#import "QCImage.h"
#import "QCImagePixelBuffer.h"

static OSStatus audioBufferHandler(AudioDeviceID		inDevice,
							   const AudioTimeStamp*	inNow,
							   const AudioBufferList*	inInputData,
							   const AudioTimeStamp*	inInputTime,
							   AudioBufferList*			outOutputData,
							   const AudioTimeStamp*	inOutputTime,
							   void*					inClientData)
{
	// we might've (accidentally) selected an audio device with no input channels
	if( inInputData->mNumberBuffers >= 1 )
	{
//		NSLog(@"read stuff! (%i buffers)", inInputData->mNumberBuffers);

		AudioInputPatch *ai = (AudioInputPatch*)inClientData;
		NSAutoreleasePool *p = [[NSAutoreleasePool alloc] init];
		[ai setAudioData:inInputData];
		[p release];
	}

	return 0;
}

@implementation AudioInputPatch : QCPatch

+ (int)executionModeWithIdentifier:(id)fp8
{
	return 2;
}
+ (BOOL)allowsSubpatchesWithIdentifier:(id)fp8
{
	return NO;
}
+ (int)timeModeWithIdentifier:(id)fp8
{
	return 1;
}

- (void)setAudioData:(const AudioBufferList *)audioBufferList
{
	QCStructure *d = [[QCStructure alloc] init], *chan;
	QCStructure *p = [[QCStructure alloc] init];
	unsigned int i, j;
	NSUInteger currentBuffer;
	NSUInteger currentChannel=0;
	for(currentBuffer=0;currentBuffer<audioBufferList->mNumberBuffers;++currentBuffer)
	{
		AudioBuffer audioBuffer = audioBufferList->mBuffers[currentBuffer];
		
		float *sampleData = (float*)audioBuffer.mData;
		float max;
		unsigned int dataSize = audioBuffer.mDataByteSize/(audioBuffer.mNumberChannels * sizeof(float));
		for(j = 0; j < audioBuffer.mNumberChannels; ++j)
		{
			NSMutableArray *channel = [(NSMutableArray*)[NSMutableArray alloc] initWithCapacity: dataSize];
			unsigned int offset = j;
			max  = fabs(sampleData[offset]);
			for(i=0; i< dataSize; ++i)
			{
				[channel addObject: [NSNumber numberWithFloat: sampleData[offset]]];
				if( fabs(sampleData[offset]) > max)
					max = fabs(sampleData[offset]);
				offset += audioBuffer.mNumberChannels;
			}
			chan = [[QCStructure alloc] initWithArray: channel];
			[channel release];
			NSString *channelNumber = [NSString stringWithFormat:@"channel%02i",currentChannel];
			[d addMember: chan forKey: channelNumber];
			[p addMember: [NSNumber numberWithFloat: max] forKey: channelNumber];
			++currentChannel;
			[chan release];
		}
	}
	@synchronized(audioData)
	{
		[audioData release];
		[audioPeaks release];
		audioData = d;
		audioPeaks = p;
	}
}

- (id)initWithIdentifier:(id)fp8
{
	if(self=[super initWithIdentifier:fp8])
	{
		//[self connect];
		pf = [QCPixelFormat pixelFormatARGB8];
	}

	return self;
}

- (id)attributes
{
	NSMutableDictionary *at=[[super attributes] mutableCopy];
	[at setObject:@"Kineme Audio Input" forKey:@"name"];
	return at;
}


- (BOOL)setup:(QCOpenGLContext *)context
{
	[self setDeviceUID:[inputDeviceUID stringValue]];

	[self connect];

	return YES;
}
- (void)cleanup:(QCOpenGLContext *)context
{
	AudioDeviceDestroyIOProcID(device, procID);
}


- (void)enable:(QCOpenGLContext *)context
{
	cs = CGColorSpaceCreateWithName(kCGColorSpaceGenericRGB);
	AudioDeviceStart(device, procID);
}
- (void)disable:(QCOpenGLContext *)context
{
	AudioDeviceStop(device, procID);
	CGColorSpaceRelease(cs);
}

- (void)connect
{
	if(device)
	{
		AudioDeviceStop(device, procID);
		AudioDeviceDestroyIOProcID(device, procID);
	}
	
	UInt32 size, count;
	
	//device = kAudioHardwarePropertyDefaultInputDevice;//devices[0];
	size = sizeof(AudioDeviceID);
	if( (deviceUID == nil) || ([deviceUID length] == 0) )
	{
		//NSLog(@"using default input device");
		AudioHardwareGetProperty(kAudioHardwarePropertyDefaultInputDevice, &size, &device);
	}
	else
	{
		NSString *uid = nil;
		unsigned int i;
		AudioHardwareGetPropertyInfo(kAudioHardwarePropertyDevices, &size, NULL);
		//NSLog(@"%i devices", size / sizeof(AudioDeviceID));
		
		count = size/sizeof(AudioDeviceID);
		AudioDeviceID devices[count];
		AudioHardwareGetProperty(kAudioHardwarePropertyDevices,&size, &devices);
		for(i=0; i < count; ++i)
		{
			size = sizeof(NSString*);
			if(AudioDeviceGetProperty(devices[i],
									  0, YES,
									  kAudioDevicePropertyDeviceUID,
									  &size, &uid))
				NSLog(@"Error getting device uid");
			if([uid isEqual: deviceUID])
			{
				//NSLog(@"found matching device %i %@", devices[i],uid);
				device = devices[i];
			}
		}
	}
	
	//AudioDeviceAddIOProc(device, audioBufferHandler, self);
	//AudioDeviceStart(device, audioBufferHandler);
	if(AudioDeviceCreateIOProcID(device,
							  audioBufferHandler,
							  self,
							  &procID))
		NSLog(@"Error creating IOProc");
	//AudioDeviceStart(device, procID);
	
	AudioStreamBasicDescription recordFormat;
	
	size = sizeof(recordFormat);
	AudioHardwareGetProperty(kAudioDevicePropertyStreamFormat, &size, &recordFormat);
	if(recordFormat.mFormatFlags & kLinearPCMFormatFlagIsFloat == FALSE)
	{
		NSLog(@"QCAudioInput:  non-floating samples not currently supported");
	}
	//NSLog(@"float samples: %i",recordFormat.mFormatFlags & kLinearPCMFormatFlagIsFloat);
	//AudioDeviceStart(device, procID);
}

- (BOOL)execute:(QCOpenGLContext *)context time:(double)time arguments:(NSDictionary *)arguments
{
	if( [inputDeviceUID wasUpdated] )
		[self setDeviceUID:[inputDeviceUID stringValue]];

	if(audioData)
	{
		unsigned int i, j, count;
		QCImagePixelBuffer *pb;
		QCStructure *localAudioData;
		QCStructure *localAudioPeaks;
		@synchronized(audioData)
		{
			localAudioData = [audioData retain];
			localAudioPeaks = [audioPeaks retain];
		}
		unsigned int channelCount = [localAudioData count];
		count = [[localAudioData memberAtIndex:0] count];
		pb = [[context imageManager] 
				createPixelBufferWithFormat: pf
				pixelsWide: count
				pixelsHigh: channelCount
				options: nil];
		unsigned int rowBytes = [pb bytesPerRow]/sizeof(unsigned int);
		unsigned int *audioImageData = (unsigned int*)[pb baseAddress];
		[pb beginUpdatePixels:FALSE colorSpace: cs];
		[outputWaveform setStructureValue: localAudioData];
		[outputPeaks setStructureValue: localAudioPeaks];
		[localAudioPeaks release];
		id audioDataMember;
		unsigned int value;
			
		for(j=0;j<channelCount;++j)
		{
			audioDataMember = [localAudioData memberAtIndex:j];
			for(i=0;i<count;++i)
			{
				value = (127.f*(1.0f+[[audioDataMember memberAtIndex: i] floatValue]));
				value *= 0x01010101;	// duplicate byte in all 4 fields
				audioImageData[i] = value;
			}
			audioImageData += rowBytes;
		}
		[localAudioData release];
		[pb endUpdatePixels];
		QCImage *audioImage = [[QCImage alloc] initWithQCImageBuffer: pb options: nil];
		[pb release];
		[outputWaveformImage setImageValue: audioImage];
		[audioImage release];
	}
	
	return YES;
}

- (void)setDeviceUID:(NSString*)uid
{
	if([uid isEqual:deviceUID] == FALSE)
	{
		[deviceUID release];
		deviceUID = [uid retain];
		[self connect];
		AudioDeviceStart(device, procID);
	}
}

-(NSString*)deviceUID
{
	return deviceUID;
}

@end
